#include <iostream>

#include "tree.h"

template <typename T>
std::ostream& operator<<(std::ostream& s, const avl_t<T>& rhs)
{
  if(rhs.empty()) s << "empty" << std::endl;
  else rhs.inorder(s,rhs.root);

  return s;
}

template <typename T>
void avl_t<T>::inorder(std::ostream& s, node_t* const &t) const
{
  if(t != NULL) {
    inorder(s,t->left);
    s << t->data << "(" << t->bal << ") " << std::endl;
    inorder(s,t->right);
  }
}

template <typename T>
avl_t<T>::avl_t()
{
  // this constructor creates an emply avl_t object
  root = NULL;
}

template <typename T>
const avl_t<T>& avl_t<T>::operator=(const avl_t<T>& rhs)
{
  if(this != &rhs) {       // standard alias test
    clear();
    root = clone(rhs.root);
  }

  return *this;
}

template <typename T>
typename avl_t<T>::node_t* avl_t<T>::clone(node_t* t) const
{
  if(t == NULL) return NULL;

  return new node_t(t->data,clone(t->left),clone(t->right),t->bal);
}

template <typename T>
void avl_t<T>::clear(node_t* &t)
{
  if(t != NULL) {
    clear(t->left);
    clear(t->right);
    delete t;
  }
  t = NULL;
}

template <typename T>
typename avl_t<T>::node_t* avl_t<T>::min(node_t* t) const
{
  return t->left == NULL ? t : min(t->left);
}

template <typename T>
typename avl_t<T>::node_t* avl_t<T>::max(node_t* t) const
{
  return t->right == NULL ? t : max(t->right);
}

template <typename T>
bool avl_t<T>::contains(const T& x, node_t* t) const
{
  if(t == NULL) return false;
  else if(x < t->data) return contains(x,t->left);
  else if(x > t->data) return contains(x,t->right);
  else return true;
}

template <typename T>
int avl_t<T>::insert(const T& x, node_t* & t)
{
	int	deltah=0;

  if(t == NULL) { t = new node_t(x,NULL,NULL);  deltah = 1; }
  else if(x < t->data) {             // go down left subtree
    if(insert(x,t->left)) {
      t->bal--;                      // height of left subtree increased
      if(t->bal == -1) deltah = 1;   // no need to rebalance here
      else if(t->bal == -2) {
        if(t->left->bal == 1) {      // if left child "right heavy"
          std::cerr << "double_";    //   rotate tree left with right child
          rotate_right(t->left);     //   (first part of double rotation)
        }
        std::cerr << "rotate_left";
        rotate_left(t);              // rotate tree right with left child
      }
    }
  } else if(x > t->data) {
    if(insert(x,t->right)) {
      t->bal++;                      // height of right subtree increased
      if(t->bal == 1) deltah = 1;    // no need to rebalance here
      else if(t->bal == 2) {
        if(t->right->bal == -1) {    // if right child "left heavy"
          std::cerr << "double_";    //   rotate tree right with left child
          rotate_left(t->right);     //   (first part of double rotation)
        }
        std::cerr << "rotate_right";
        rotate_right(t);             // rotate tree left with right child
      }
    }
  }
  else; // duplicate, do nothing

  return deltah;
}

template <typename T>
void avl_t<T>::rotate_left(node_t* & p)
{
  // rotate tree node with left child
  // for AVL trees, this is single rotation for case 1
  // set new root then update balance info
  node_t* q = p;
  p = p->left;                       // root is rotated up from below left
  q->left = p->right;                // tree is rotated once to the right
  p->right = q;
  q->bal++;                          // update balance info
  if(p->bal < 0) q->bal -= p->bal;
  p->bal++;
  if(q->bal > 0) p->bal += q->bal;
}

template <typename T>
void avl_t<T>::rotate_right(node_t* & p)
{
  // rotate tree node with right child
  // for AVL trees, this is single rotation for case 4
  // set new root then update balance info
  node_t* q = p;
  p = p->right;                      // root is rotated up from below right
  q->right = p->left;                // tree is rotated once to the left
  p->left = q;
  q->bal--;                          // update balance info
  if(p->bal > 0) q->bal -= p->bal;
  p->bal--;
  if(q->bal < 0) p->bal += q->bal;
}

template <typename T>
int avl_t<T>::erase(const T& x, node_t* & t)
{
        int deltah=0;

  if(t == NULL) return 0; // item not found, do nothing

  if(x < t->data) {
    if(erase(x,t->left)) {
      // almost like insert(x,t->right) above
      t->bal++;                      // height of left subtree decreased
      if(t->bal == 0) deltah = 1;    // may need to rebalance above
      else if(t->bal == 2) {
        if(t->right->bal == -1) {    // if right child "left heavy"
          std::cerr << "double_";    //   rotate tree right with left child
          rotate_left(t->right);     //   (first part of double rotation)
        }
        std::cerr << "rotate_right ";
        rotate_right(t);             // rotate tree left with right child
        if(t->bal == 0) deltah = 1;  // may need to rebalance above
      }
    }
  } else if(x > t->data) {
    if(erase(x,t->right)) {
      // almost like insert(x,t->left) above
      t->bal--;                      // height of right subtree decreased
      if(t->bal == 0) deltah = 1;    // may need to rebalance above
      else if(t->bal == -2) {
        if(t->left->bal == 1) {      // if left child "right heavy"
          std::cerr << "double_";    //   rotate tree left with right child
          rotate_right(t->left);     //   (first part of double rotation)
        }
        std::cerr << "rotate_left ";
        rotate_left(t);              // rotate tree right with left child
        if(t->bal == 0) deltah = 1;  // may need to rebalance above
      }
    }
  } else if(t->left != NULL && t->right != NULL) { // two children
    t->data = min(t->right)->data;
    if(erase(t->data,t->right)) {
      // almost like insert(x,t->left) above
      t->bal--;                      // height of right subtree decreased
      if(t->bal == 0) deltah = 1;    // may need to rebalance above
      else if(t->bal == -2) {
        if(t->left->bal == 1) {      // if left child "right heavy"
          std::cerr << "double_";    //   rotate tree left with right child
          rotate_right(t->left);     //   (first part of double rotation)
        }
        std::cerr << "rotate_left ";
        rotate_left(t);              // rotate tree right with left child
        if(t->bal == 0) deltah = 1;  // may need to rebalance above
      }
    }
  } else {
        node_t* old = t;
    t = (t->left != NULL) ? t->left : t->right;
    delete old;
    deltah = 1;
  }
 
  return deltah;
}

/////////////////////////// specializations ////////////////////////////////////
template class avl_t<int>;
template std::ostream& operator<<(std::ostream&, const avl_t<int>&);

template class avl_t<float>;
template std::ostream& operator<<(std::ostream&, const avl_t<float>&);

template class avl_t<double>;
template std::ostream& operator<<(std::ostream&, const avl_t<double>&);
